//_________________________________________________________________________
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
//$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// NOMBRE: GBASystem.c
// AUTOR: Eloist and Eduardo Pinera Aznarez
// DESCRIPCION: Registros, Variables y funciones genericas de la 
//	Game Boy Advanced.
//
// FECHA INICIO: 3 de Noviembre de 2000.
// ULTIMA ACTUALIZACION: 16 de Noviembre de 2000.
//$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
//

#ifndef GBASYSTEM_HEADER
#define GBASYSTEM_HEADER

#include "stdtypes.h"

#define	TRUE			0
#define	FALSE			1

#define	SUCCEDED		0
#define	FAILED			1

#define S_OK			0
#define E_FAIL			1

typedef void			VOID;

typedef unsigned char	BYTE;
typedef unsigned short	WORD;
typedef unsigned long	DWORD;

typedef signed char		CHAR;
typedef signed short	INT;
typedef signed long		LONG;

#define BG_MODE_0 0
#define BG_MODE_1 1
#define BG_MODE_2 2
#define BG_MODE_3 3
#define BG_MODE_4 4
#define BG_MODE_5 5

#define BG0_ENABLE 1
#define BG1_ENABLE 2
#define BG2_ENABLE 4
#define BG3_ENABLE 8
#define OBJ_ENABLE 16

#define OBJ_1D_MAP 1
#define OBJ_2D_MAP 0

#define CHR_BASE 0xC

#define REG_DISPCNT    *(unsigned long*)0x4000000
#define REG_DISPCNT_L  *(unsigned short*)0x4000000
#define REG_DISPCNT_H  *(unsigned short*)0x4000002
#define REG_DISPSTAT   *(unsigned short*)0x4000004
#define REG_VCOUNT     *(unsigned short*)0x4000006
#define REG_BG0CNT     *(unsigned short*)0x4000008
#define REG_BG1CNT     *(unsigned short*)0x400000A
#define REG_BG2CNT     *(unsigned short*)0x400000C
#define REG_BG3CNT     *(unsigned short*)0x400000E
#define REG_BG0HOFS    *(unsigned short*)0x4000010
#define REG_BG0VOFS    *(unsigned short*)0x4000012
#define REG_BG1HOFS    *(unsigned short*)0x4000014
#define REG_BG1VOFS    *(unsigned short*)0x4000016
#define REG_BG2HOFS    *(unsigned short*)0x4000018
#define REG_BG2VOFS    *(unsigned short*)0x400001A
#define REG_BG3HOFS    *(unsigned short*)0x400001C
#define REG_BG3VOFS    *(unsigned short*)0x400001E
#define REG_BG2PA      *(unsigned short*)0x4000020
#define REG_BG2PB      *(unsigned short*)0x4000022
#define REG_BG2PC      *(unsigned short*)0x4000024
#define REG_BG2PD      *(unsigned short*)0x4000026
#define REG_BG2X       *(unsigned long*)0x4000028
#define REG_BG2X_L     *(unsigned short*)0x4000028
#define REG_BG2X_H     *(unsigned short*)0x400002A
#define REG_BG2Y       *(unsigned long*)0x400002C
#define REG_BG2Y_L     *(unsigned short*)0x400002C
#define REG_BG2Y_H     *(unsigned short*)0x400002E
#define REG_BG3PA      *(unsigned short*)0x4000030
#define REG_BG3PB      *(unsigned short*)0x4000032
#define REG_BG3PC      *(unsigned short*)0x4000034
#define REG_BG3PD      *(unsigned short*)0x4000036
#define REG_BG3X       *(unsigned long*)0x4000038
#define REG_BG3X_L     *(unsigned short*)0x4000038
#define REG_BG3X_H     *(unsigned short*)0x400003A
#define REG_BG3Y       *(unsigned long*)0x400003C
#define REG_BG3Y_L     *(unsigned short*)0x400003C
#define REG_BG3Y_H     *(unsigned short*)0x400003E
#define REG_WIN0H      *(unsigned short*)0x4000040
#define REG_WIN1H      *(unsigned short*)0x4000042
#define REG_WIN0V      *(unsigned short*)0x4000044
#define REG_WIN1V      *(unsigned short*)0x4000046
#define REG_WININ      *(unsigned short*)0x4000048
#define REG_WINOUT     *(unsigned short*)0x400004A
#define REG_MOSAIC     *(unsigned long*)0x400004C
#define REG_MOSAIC_L   *(unsigned long*)0x400004C
#define REG_MOSAIC_H   *(unsigned long*)0x400004E
#define REG_BLDMOD     *(unsigned short*)0x4000050
#define REG_COLEY1     *(unsigned short*)0x4000052
#define REG_COLEY2     *(unsigned short*)0x4000054
#define REG_SG10       *(unsigned long*)0x4000060
#define REG_SG10_L     *(unsigned short*)0x4000060
#define REG_SG10_H     *(unsigned short*)0x4000062
#define REG_SG11       *(unsigned short*)0x4000064
#define REG_SG20       *(unsigned short*)0x4000068
#define REG_SG21       *(unsigned short*)0x400006C
#define REG_SG30       *(unsigned long*)0x4000070
#define REG_SG30_L     *(unsigned short*)0x4000070
#define REG_SG30_H     *(unsigned short*)0x4000072
#define REG_SG31       *(unsigned short*)0x4000074
#define REG_SG40       *(unsigned short*)0x4000078
#define REG_SG41       *(unsigned short*)0x400007C
#define REG_SGCNT0     *(unsigned long*)0x4000080
#define REG_SGCNT0_L   *(unsigned short*)0x4000080
#define REG_SGCNT0_H   *(unsigned short*)0x4000082
#define REG_SGCNT1     *(unsigned short*)0x4000084
#define REG_SGBIAS     *(unsigned short*)0x4000088
#define REG_SGWR0      *(unsigned long*)0x4000090
#define REG_SGWR0_L    *(unsigned short*)0x4000090
#define REG_SGWR0_H    *(unsigned short*)0x4000092
#define REG_SGWR1      *(unsigned long*)0x4000094
#define REG_SGWR1_L    *(unsigned short*)0x4000094
#define REG_SGWR1_H    *(unsigned short*)0x4000096
#define REG_SGWR2      *(unsigned long*)0x4000098
#define REG_SGWR2_L    *(unsigned short*)0x4000098
#define REG_SGWR2_H    *(unsigned short*)0x400009A
#define REG_SGWR3      *(unsigned long*)0x400009C
#define REG_SGWR3_L    *(unsigned short*)0x400009C
#define REG_SGWR3_H    *(unsigned short*)0x400009E
#define REG_SGFIF0A    *(unsigned long*)0x40000A0
#define REG_SGFIFOA_L  *(unsigned short*)0x40000A0
#define REG_SGFIFOA_H  *(unsigned short*)0x40000A2
#define REG_SGFIFOB    *(unsigned long*)0x40000A4
#define REG_SGFIFOB_L  *(unsigned short*)0x40000A4
#define REG_SGFIFOB_H  *(unsigned short*)0x40000A6
#define REG_DMA0SAD     *(unsigned long*)0x40000B0
#define REG_DMA0SAD_L   *(unsigned short*)0x40000B0
#define REG_DMA0SAD_H   *(unsigned short*)0x40000B2
#define REG_DMA0DAD     *(unsigned long*)0x40000B4
#define REG_DMA0DAD_L   *(unsigned short*)0x40000B4
#define REG_DMA0DAD_H   *(unsigned short*)0x40000B6
#define REG_DMA0CNT     *(unsigned long*)0x40000B8
#define REG_DMA0CNT_L   *(unsigned short*)0x40000B8
#define REG_DMA0CNT_H   *(unsigned short*)0x40000BA
#define REG_DMA1SAD     *(unsigned long*)0x40000BC
#define REG_DMA1SAD_L   *(unsigned short*)0x40000BC
#define REG_DMA1SAD_H   *(unsigned short*)0x40000BE
#define REG_DMA1DAD     *(unsigned long*)0x40000C0
#define REG_DMA1DAD_L   *(unsigned short*)0x40000C0
#define REG_DMA1DAD_H   *(unsigned short*)0x40000C2
#define REG_DMA1CNT     *(unsigned long*)0x40000C4
#define REG_DMA1CNT_L   *(unsigned short*)0x40000C4
#define REG_DMA1CNT_H   *(unsigned short*)0x40000C6
#define REG_DMA2SAD     *(unsigned long*)0x40000C8
#define REG_DMA2SAD_L   *(unsigned short*)0x40000C8
#define REG_DMA2SAD_H   *(unsigned short*)0x40000CA
#define REG_DMA2DAD     *(unsigned long*)0x40000CC
#define REG_DMA2DAD_L   *(unsigned short*)0x40000CC
#define REG_DMA2DAD_H   *(unsigned short*)0x40000CE
#define REG_DMA2CNT     *(unsigned long*)0x40000D0
#define REG_DMA2CNT_L   *(unsigned short*)0x40000D0
#define REG_DMA2CNT_H   *(unsigned short*)0x40000D2
#define REG_DMA3SAD     *(unsigned long*)0x40000D4
#define REG_DMA3SAD_L   *(unsigned short*)0x40000D4
#define REG_DMA3SAD_H   *(unsigned short*)0x40000D6
#define REG_DMA3DAD     *(unsigned long*)0x40000D8
#define REG_DMA3DAD_L   *(unsigned short*)0x40000D8
#define REG_DMA3DAD_H   *(unsigned short*)0x40000DA
#define REG_DMA3CNT     *(unsigned long*)0x40000DC
#define REG_DMA3CNT_L   *(unsigned short*)0x40000DC
#define REG_DMA3CNT_H   *(unsigned short*)0x40000DE
#define REG_TM0D       *(unsigned short*)0x4000100
#define REG_TM0CNT     *(unsigned short*)0x4000102
#define REG_TM1D       *(unsigned short*)0x4000104
#define REG_TM1CNT     *(unsigned short*)0x4000106
#define REG_TM2D       *(unsigned short*)0x4000108
#define REG_TM2CNT     *(unsigned short*)0x400010A
#define REG_TM3D       *(unsigned short*)0x400010C
#define REG_TM3CNT     *(unsigned short*)0x400010E
#define REG_SCD0       *(unsigned short*)0x4000120
#define REG_SCD1       *(unsigned short*)0x4000122
#define REG_SCD2       *(unsigned short*)0x4000124
#define REG_SCD3       *(unsigned short*)0x4000126
#define REG_SCCNT      *(unsigned long*)0x4000128
#define REG_SCCNT_L    *(unsigned short*)0x4000128
#define REG_SCCNT_H    *(unsigned short*)0x400012A
#define REG_P1         *(unsigned short*)0x4000130
#define REG_P1CNT      *(unsigned short*)0x4000132
#define REG_R          *(unsigned short*)0x4000134
#define REG_HS_CTRL    *(unsigned short*)0x4000140
#define REG_JOYRE      *(unsigned long*)0x4000150
#define REG_JOYRE_L    *(unsigned short*)0x4000150
#define REG_JOYRE_H    *(unsigned short*)0x4000152
#define REG_JOYTR      *(unsigned long*)0x4000154
#define REG_JOYTR_L    *(unsigned short*)0x4000154
#define REG_JOYTR_H    *(unsigned short*)0x4000156
#define REG_JSTAT      *(unsigned long*)0x4000158
#define REG_JSTAT_L    *(unsigned short*)0x4000158
#define REG_JSTAT_H    *(unsigned short*)0x400015A
#define REG_IE         *(unsigned short*)0x4000200
#define REG_IF         *(unsigned short*)0x4000202
#define REG_WSCNT      *(unsigned short*)0x4000204
#define REG_IME        *(unsigned short*)0x4000208
#define REG_PAUSE      *(unsigned short*)0x4000300


#define PBYTE_PALETTE	((unsigned char  *) 0x05000000)
#define PWORD_PALETTE	((unsigned short *) 0x05000000)
#define PDWORD_PALETTE	((unsigned long  *) 0x05000000)

#define PBYTE_SCREEN	((unsigned char  *) 0x06000000)
#define PWORD_SCREEN	((unsigned short *) 0x06000000)
#define PDWORD_SCREEN	((unsigned long  *) 0x06000000)

#define PBYTE_TILES  ((u8*) 0x6004000)
#define PWORD_TILES  ((u16*) 0x6004000)
#define PDWORD_TILES  ((u32*) 0x6004000)

#define OBJ_PLTT	((u16*) 0x05000200)
#define OBJCHARDATA_MODE0_2	((u8*) 0x06010000)
#define OBJCHARDATA_MODE3_5   ((u8*) 0x06014000)
#define PWORD_OBJCHARDATA_MODE3_5   ((u16*) 0x06014000)

//Strutura dell'oam
typedef struct _OAMSprite { u16 Attrib0, Attrib1, Attrib2, RotateZoom; } OAMSprite;

//Puntatore all'area dell'oam
#define OAM ((OAMSprite*)0x07000000)

extern u32 g_VCount;
/************************/
/* Functions Prototypes */
/************************/
void WaitVerticalRetrace(void);
void DMA3Call(u32 Src, u32 Dst, u32 Cnt);

#endif	//GBASYSTEM_HEADER