/* Applicazione di test per il gameboy advance */

/*
//_________________________________________________________________________
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
//$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// NOMBRE: GBAMain.c
// AUTOR: Eduardo Pinera Aznarez
// HACKED by : Francesco Pischedda :)
// DESCRIPCION: Entrada del programa.
//
// FECHA INICIO: 3 de Noviembre de 2000.
// ULTIMA ACTUALIZACION: 19/04/2001
//$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
//
*/

#include <stdio.h>
#include <stdlib.h>

#include "GBASystem.h"
#include "GBADDraw.h"
#include "GBADInput.h"
#include "sprite.h"
#include "defPal.h"
#include "sfondo.h"
#include "lecca.h"
#include "stecca.h"

//#define INTRO
#undef INTRO

#ifdef INTRO
#include "introback.h"
#endif


#define MARGIN_LEFT 8
#define MARGIN_RIGHT 72
#define MARGIN_TOP 8
#define MARGIN_BOTTOM 136

#undef USEMODE4
//#define USEMODE4


#undef BACKROTATION
//#define BACKROTATION

enum {REDBLOCK=0,BLUEBLOCK,GREENBLOCK,YELLOWBLOCK,VIOLETBLOCK,REDSTICK,BLUESTICK,GREENSTICK,YELLOWSTICK,VIOLETSTICK,NOBLOCK};

#define ARENAWIDTH 5
#define ARENAHEIGHT 8

void screen_show(u8 *image, u16 *colormap);
void dma_screen_show(u8 *image, u16 *colormap);
void drawObj(u32 r, u32 c, u32 type);
void drawLecca(u32 x, u32 y, u32 col);
void drawLecca16x16(u32 x, u32 y, u32 col);
void drawStecca16x16(u32 x, u32 y, u32 col);
void drawArena(u32 x, u32 y, u8 a[ARENAHEIGHT][ARENAWIDTH]);
int cellFree(u32 x,u32 y,u8 a[ARENAHEIGHT][ARENAWIDTH], u8 type);
int checkLecca(int32 r, int32 c, u32 type, u8 a[ARENAHEIGHT][ARENAWIDTH]);
int checkStick(int32 r, int32 c, u32 type, u8 a[ARENAHEIGHT][ARENAWIDTH]);
void checkSimilar(u32 r, u32 c, u32 type, u8 a[ARENAHEIGHT][ARENAWIDTH]);
void delCheckedObj(u8 a[ARENAHEIGHT][ARENAWIDTH]);
void delObj(u32 r, u32 c);
void resetCheckTmp();

void Intro();
void InitGame();

void initArena(u8 a[ARENAHEIGHT][ARENAWIDTH]){

	u8 i,j;

	for(i=0;i<ARENAHEIGHT;++i)
		for(j=0;j<ARENAWIDTH;++j)
			a[i][j]=NOBLOCK;
}

OAMSprite *g_oam; //puntaotore globale all'oam
u8 checkTmp[ARENAWIDTH];//matice di che indica se una casella e' gia stata controllata

#define ISDEL(r,c) (checkTmp[c] & (1<<r))
#define SETDELETED(r,c) (checkTmp[(c)] |= (1<<(r)))
#define UNSETDELETED(r,c) (checkTmp[(c)] &= (!(1<<(r))))

u8 deletedObj[ARENAHEIGHT*ARENAWIDTH][2];//vettore contenente le posizioni degli oggetti cancellati
u32 delPosition;

#define InitScene() g_oam = OAM

void GBAMain(void)
{

   u16 PadState;
   u16 x,y;
	u8 type;
	u8 arena1[ARENAHEIGHT][ARENAWIDTH],arena2[ARENAHEIGHT][ARENAWIDTH];//le arene di gioco
   
   // Mostro l'intro
#ifdef INTRO
   Intro();

   //Aspetto che venga premuto start
   ClearPad();
   do
      ReadPad(&PadState);
   while(PadState!=KEY_START);
   ClearPad();
#endif
   
   /*Inizializzo lo schermo nel modo 0*/
   /*Carico immagini e palette*/
   InitGame();

   x=40;
   y=16;
	srand(*((u16*)0x04000006));
   
	type = rand()%10;

	//init the arena matrixes
	initArena(arena1);initArena(arena2);
   
   #ifdef BACKROTATION
      xr = yr = 0;
      xa = ya = 0;
   #endif


   while(1)
	{
		ReadPad(&PadState);

		switch(PadState)
		{
			case KEY_LEFT:
				if(x!=MARGIN_LEFT)
               x -= 16;
				ClearPad();
				break;

			case KEY_UP:
				//sPalla.y -= 3;
				ClearPad();
				break;

			case KEY_RIGHT:
				if(x!=MARGIN_RIGHT) 
               x += 16;
				ClearPad();
				break;

			case KEY_DOWN:
				y += 3;
				ClearPad();
				break;
		};

      y+=1;
      if(y>MARGIN_BOTTOM){
			//assegno il leccalecca alla matrice dei leccalecca
			arena1[ARENAHEIGHT-1][(x-8)/16] = type;
         drawObj(ARENAHEIGHT-1,(x-8)/16,type);
         checkSimilar(ARENAHEIGHT-1,(x-8)/16,type,arena1);
			y=MARGIN_TOP;
         x=40;
			type = rand()%10;
		}
		else{
			if(!cellFree(x-8,y-8,arena1,type)){
            checkSimilar((y-8)/16-1,(x-8)/16,type,arena1);
				y=MARGIN_TOP;
            x=40;
				type = rand()%10;
			}
		}

      WaitVerticalRetrace();

		InitScene();

      /*Disegno il leccalecca*/
      if(type<REDSTICK)
         drawLecca16x16(x,y,type);
      else
         drawStecca16x16(x,y,type-REDSTICK);

      //debug
      /*
      drawObj(3,0,NOBLOCK);
      drawObj(3,1,NOBLOCK);
      drawObj(3,2,NOBLOCK);
      drawObj(3,3,NOBLOCK);
      drawObj(3,4,NOBLOCK);
      */

		//disegno l'arena
		//drawArena(8,8,arena1);
	}
}



void screen_show(u8 *image, u16 *colormap)
{
  int i;
  u16 *pal;
  u8  *video;

  /* Set graphics mode.*/
  REG_DISPCNT = 0x0404;

  /* Set up pointers */

  pal   = PWORD_PALETTE;
  video = PBYTE_SCREEN;

  /* Set up the palette ... */

  for (i=0; i<255; i++)
  {
    *(pal+i) = *(colormap+i);
  }

  /* ... and image. */

  for (i=0; i<(160*240); i++)
  {
    *(video+i) = *(image+i);
  }

}

void dma_screen_show(u8 *image, u16 *colormap)
{
  /* Set graphics mode. */
  REG_DISPCNT = 0x1444;

  /* Set up the palette ... */
  DMA3Call((u32)colormap, (u32)PWORD_PALETTE, 0x84000100 );

  /* ... and image. */
  DMA3Call((u32)image, (u32)PWORD_SCREEN, 0x84000000 + (240*160));
}

/* mostra l'introduzione */
#ifdef INTRO
void Intro(){

   int i,j;

   // Sets video mode to 0 with BG0 and OBJ on. 1 Dimensional OBJ mode
   REG_DISPCNT = 0x1140;
   //set a virtual screen of 256x256 chars
   //REG_BG0CNT = 0x0083;
   REG_BG0CNT = 0x0084;


   //setto la palette del background
   //load the pallette info into the desired address
   DMA3Call((u32)introBackPal, (u32)PWORD_PALETTE, 0x84000100);

   //Metto l'immagine di sfondo nella zona di memoria dedicata alle tiles
   //put the background image in to the tile's memory
   DMA3Call((u32)introBackData, (u32)PBYTE_TILES, 0x84000000 + introBackDataLen/2);

   //Imposto la mappa in modo da disegnare l'immagine di sfondo
   //set up the map to display the backgroung image
   for(i=0;i<20;++i)
      for (j=0;j<30;++j) 
		   PWORD_SCREEN[i*32+j]=introBackMap[i*30+j];
}
#endif

/* Inizializza il game boy per il gioco */
void InitGame(void)
{

   int i,j;


   //#ifndef BACKROTATION
      // Sets video mode to 0 with BG0 and OBJ on. 1 Dimensional OBJ mode
      REG_DISPCNT = 0x1140;
/*   #else
      // Sets video mode to 1 with BG0,BG2 and OBJ on. 1 Dimensional OBJ mode
      REG_DISPCNT = 0x1541;

      //set a virtual screen screen of 256 256 chars
      REG_BG2CNT = 0x1144;
   #endif*/
   //set a virtual screen of 256x256 chars
   REG_BG0CNT = 0x0084;


   //setto la palette del background
   //load the pallette info into the desired address
   DMA3Call((u32)backPal, (u32)PWORD_PALETTE, 0x84000100);

   //Metto l'immagine di sfondo nella zona di memoria dedicata alle tiles
   //put the background image in to the tile's memory
   DMA3Call((u32)backData, (u32)PBYTE_TILES, 0x84000000 + backDataLen/2);

   //Imposto la mappa in modo da disegnare l'immagine di sfondo
   //set up the map to display the backgroung image
   for(i=0;i<20;++i)
      for (j=0;j<30;++j) 
		   PWORD_SCREEN[i*32+j]=backMap[i*30+j];

   //resetto L'OAM
   for(i=0;i<1024;++i){
      OAM[i].Attrib0 = 180;
      OAM[i].Attrib1 = 290;
      OAM[i].Attrib2 = 0;
   }


   //Setto i dati della palette e le immagini degli oggetti
	DMA3Call((u32)objPal, (u32)OBJ_PLTT, 0x84000100);	// Setup the sprites palette

   //Immagine del leccalecca
   DMA3Call((u32)leccaData, (u32)OBJCHARDATA_MODE0_2, 0x84000000 + leccaDataLen/2);		// Copy Sprites Character Bloc to VRAM

   //Immagine della stecca
   DMA3Call((u32)steccaData, (u32)OBJCHARDATA_MODE0_2+leccaDataLen, 0x84000000 + leccaDataLen/2);		// Copy Sprites Character Bloc to VRAM

   //Resetto la matrice del controllo
   for(i=0;i<ARENAWIDTH;++i)
      checkTmp[i]=0;
}

/* Associa una cella ad una posizione nell'oam */
void drawObj(u32 r, u32 c, u32 type){

   u16 pal;
   register u32 x = 8+(c<<4), y= 24 + (r<<4);
   register OAMSprite* oam_ptr = OAM + 2 + c + r*ARENAWIDTH;

   if(type==NOBLOCK){
      oam_ptr->Attrib0 = 180;
      oam_ptr->Attrib1 = 290;
      oam_ptr->Attrib2 = 0;
   }
   else if(type>=REDSTICK){
      //disegno l'immagine della stecchetta in base al suo colore
      pal = (type-REDSTICK)<<12;
      oam_ptr->Attrib0 = y;
      oam_ptr->Attrib1 = 0x4000 | x;
      oam_ptr->Attrib2 = pal+4;
   }
   else{
      //disegno l'immagine del leccalecca in base al colore
      pal = type<<12;
      oam_ptr->Attrib0 = y;
      oam_ptr->Attrib1 = 0x4000 | x;
      oam_ptr->Attrib2 = pal;
   }
}

/* disegna un leccalecca fatto di 4 tile 8x8*/
void drawLecca(u32 x, u32 y, u32 col){

   //uso il colore per selezionare la palette
   register u16 pal = col << 12;

   //disegno l'immagine del leccalecca in base al colore
   g_oam->Attrib0 = y;
   g_oam->Attrib1 = x;
   g_oam->Attrib2 = pal;
   ++g_oam;

   g_oam->Attrib0 = y;
   g_oam->Attrib1 = x+8;
   g_oam->Attrib2 = pal+1;
   ++g_oam;y+=8;

   g_oam->Attrib0 = y;
   g_oam->Attrib1 = x;
   g_oam->Attrib2 = pal+2;
   ++g_oam;

   g_oam->Attrib0 = y;
   g_oam->Attrib1 = x+8;
   g_oam->Attrib2 = pal+3;
	++g_oam;
}

/* Disegna un leccalecca fatto di una tile 32x32 */
void drawLecca16x16(u32 x, u32 y, u32 col){

   //uso il colore per selezionare la palette
   register u16 pal = col <<12;

   //disegno l'immagine del leccalecca in base al colore
   g_oam->Attrib0 = y;
   g_oam->Attrib1 = 0x4000 | x;
   g_oam->Attrib2 = pal;
   ++g_oam;
}

/* Disegna una stecchetta fatta da una tile 16x16 */
void drawStecca16x16(u32 x, u32 y, u32 col){

   //uso il colore per selezionare la palette
   register u16 pal = col <<12;

   //disegno l'immagine del leccalecca in base al colore
   g_oam->Attrib0 = y;
   g_oam->Attrib1 = 0x4000 | x;
   g_oam->Attrib2 = pal + 4;
   ++g_oam;
}

/* Disegna l'arena a partire dalle coordinate x,y */
/* NOTA : l'arena viene disegnata per colonne a partire dall'ultima riga              */
/* in questo modo ci si pu fermare quando non ci sono pi leccalecca in una          */
/* colonna evitando controlli inutili. Un altro sistema migliore potrebbe essere      */
/* quello di mmorizzare il numero di leccalecca in una colonna e di fare n iterazioni */
/* per quanti sono i LL nella colonna.                                                */
void drawArena(u32 x, u32 y, u8 a[ARENAHEIGHT][ARENAWIDTH]){

	int8 i,j;
	u8 *a_ptr;

	for(i=0;i<ARENAWIDTH;++i){

		a_ptr=(u8*)&a[ARENAHEIGHT-1][i];

		for(j=ARENAHEIGHT-1;j>=0;--j){

         drawObj(j,i,*a_ptr);

			a_ptr-=ARENAWIDTH;
		}
	}
}

/* Indica se la cella alle coordinate x,y dell'arena  libera*/
int cellFree(u32 x, u32 y, u8 a[ARENAHEIGHT][ARENAWIDTH], u8 type){

	y>>=4;
	x>>=4;
	if(a[y][x]!=NOBLOCK){
		a[y-1][x] = type;
      drawObj(y-1,x,type);//attivo l'oggetto corrispettivo nell'oam
		return 0;
	}

	return 1;
}

void drawCol(u8 a[ARENAHEIGHT][ARENAWIDTH], u32 col){

   u32 i;

   for(i=0;i<ARENAHEIGHT;++i)
      drawObj(i,col,a[i][col]);
}

void compactCol(u8 a[ARENAHEIGHT][ARENAWIDTH], u32 col){

   int32 f_free=ARENAHEIGHT-1,f_block;

   //cerco il primo blocco vuoto
   while(a[f_free][col]!=NOBLOCK && f_free-->=0) ;

   f_block = f_free-1;

   //se la colonna  tutta piena esce
   while(f_free>=0){

      //ricerca del pimo blocco non vuoto
      while(f_block>=0 && a[f_block][col]==NOBLOCK) --f_block;

      //non si sono trovati blocchi non vuoti sopra quelli vuoti
      if(f_block<0) return;

      a[f_free][col] = a[f_block][col];
      a[f_block][col] = NOBLOCK;

      --f_free;
      --f_block;
   }
}


/* Controlla se nella casella in r,c c' un oggetto di tipo type */
/* ritorna quanti oggetti dello stesso tipo sono stati trovati   */
int checkLecca(int32 r, int32 c, u32 type, u8 a[ARENAHEIGHT][ARENAWIDTH]){

   //se la riga e la colonna sono fuori dai limiti esci
   if(r>=ARENAHEIGHT || c>=ARENAWIDTH || r<0 || c<0) return 0;

   if(ISDEL(r,c)) return 0;//casella gi controllata

   if(a[r][c] == type){//L'oggetto  del tipo cercato, cotrollo intorno se ci sono altri dello stesso genere
   
      delObj(r,c);//memorizzo i dati per cancellare l'oggetto

      //controlla se intorno ci sono leccalecca e se sopra e sotto ci sono stick
      return 1 + checkLecca(r,c+1,type,a) + checkLecca(r+1,c,type,a) +
            checkLecca(r,c-1,type,a) + checkLecca(r-1,c,type,a) +
            checkStick(r+1,c,type+REDSTICK,a) + checkStick(r-1,c,type+REDSTICK,a);
   }

   return 0;
}

/* Controlla se nella casella in r,c c' un oggetto di tipo type */
/* ritorna quanti oggetti dello stesso tipo sono stati trovati   */
int checkStick(int32 r, int32 c, u32 type, u8 a[ARENAHEIGHT][ARENAWIDTH]){

   //se la riga e la colonna sono fuori dai limiti esci
   if(r>=ARENAHEIGHT || c>=ARENAWIDTH || r<0 || c<0) return 0;

   if(ISDEL(r,c)) return 0;//casella gi controllata

   if(a[r][c] == type){//L'oggetto  del tipo cercato, cotrollo intorno se ci sono altri dello stesso genere
   
      delObj(r,c);//memorizzo i dati per cancellare l'oggetto

      //controlla se sopra e sotto ci sono leccalecca e stick
      return 1 + checkStick(r+1,c,type,a) + checkStick(r-1,c,type,a) +
            checkLecca(r+1,c,type-REDSTICK,a) + checkLecca(r-1,c,type-REDSTICK,a);
   }

   return 0;
}

/* Controlla che nei dintorni dell'oggetto ci siano suoi simili */
void checkSimilar(u32 r, u32 c, u32 type, u8 a[ARENAHEIGHT][ARENAWIDTH]){

   delPosition = 0;
   int i;

   if(type<REDSTICK){

      //Controllo se sotto ho uno stick dello stesso colore del leccalecca
      if(r<ARENAHEIGHT-1 && a[r+1][c]==type+REDSTICK){
         if(checkLecca(r,c,type,a)>1){

            delCheckedObj(a);//Elimina gli oggetti cancellati dalla matrice
            for(i=0;i<ARENAWIDTH;++i){
               //Compatto le colonne
               if(checkTmp[i]){
                  compactCol(a,i);
                  drawCol(a,i);
               }
            }
            //resetto la matrice temporanea
            resetCheckTmp();
         }
         else{
            UNSETDELETED(r,c);//resetto la matrice di bit per il controllo
         }
      }
   }
   else{//controllo se sotto ho un leccalecca del colore della stecca
      if(r<ARENAHEIGHT-1 && a[r+1][c]==type-REDSTICK){
         if(checkStick(r,c,type,a)>1){

            delCheckedObj(a);//Elimina gli oggetti cancellati dalla matrice
            //Compatto le colonne
            for(i=0;i<ARENAWIDTH;++i){
               //Compatto le colonne
               if(checkTmp[i]){
                  compactCol(a,i);
                  drawCol(a,i);
               }
            }

            //resetto la matrice di bit temporanea
            resetCheckTmp();
         }
         else{
            UNSETDELETED(r,c);//resetto la matrice di bit per il controllo
         }
      }
   }
}


/* cancella un oggetto */
void delObj(u32 r, u32 c){

   SETDELETED(r,c);//dico che questo oggetto  gi stato controllato

   deletedObj[delPosition][0] = r;
   deletedObj[delPosition][1] = c;

   ++delPosition;
}

/* Cancella gli oggetti dalla matrice */
void delCheckedObj(u8 a[ARENAHEIGHT][ARENAWIDTH]){

   u32 i;
   register u32 r,c;

   for(i=0;i<delPosition;++i){
   
      r = deletedObj[i][0];
      c = deletedObj[i][1];
      a[r][c] = NOBLOCK;
      //UNSETDELETED(r,c);
   }
}

void resetCheckTmp(){

   checkTmp[0]=0;
   checkTmp[1]=0;
   checkTmp[2]=0;
   checkTmp[3]=0;
   checkTmp[4]=0;
}